from pathlib import Path

import MDAnalysis as mda
from textual.validation import (
    Failure,
    Integer,
    Number,
    ValidationResult,
    Validator,
)


class NumberOrNoneValidator(Number):
    """Check that input is empty or is a number"""

    def __init__(self, failure_description):
        super().__init__(failure_description=failure_description)

    def validate(self, value: str) -> ValidationResult:
        value = value if value else "0"
        return super().validate(value)


class IntegerOrNoneValidator(Integer):
    """Check that input is empty or can be cast to an integer"""

    def __init__(self, failure_description):
        super().__init__(failure_description=failure_description)

    def validate(self, value: str) -> ValidationResult:
        value = value if value else "0"
        return super().validate(value) if "." not in value else self.failure()


class FileValidator(Validator):
    """Check that input is a valid file."""

    def __init__(
        self,
        failure_description: str | None = None,
    ) -> None:
        super().__init__(failure_description=failure_description)

    class InvalidFile(Failure):
        """Indicates that the file does not exist."""

    def validate(self, value: str) -> ValidationResult:
        """Validates that `value` is a valid file (i.e. it exists).

        Args:
            value: The value to validate.

        Returns:
            The result of the validation.
        """
        invalid_file = ValidationResult.failure([FileValidator.InvalidFile(self, value)])
        file = Path(value).resolve()
        if not file.exists() or not file.is_file():
            return invalid_file
        return self.success()


class FileExtensionValidator(Validator):
    """Check that input has a valid file extension."""

    def __init__(
        self,
        failure_description: str | None = None,
        valid_extensions: list[str | Path] | None = None,
    ) -> None:
        super().__init__(failure_description=failure_description)
        self.valid_extensions = valid_extensions

    class InvalidFileExtension(Failure):
        """Indicates that the file extension is invalid."""

    def validate(self, value: str) -> ValidationResult:
        """Validates that `value` is a valid file extension.

        Args:
            value: The value to validate.

        Returns:
            The result of the validation.
        """
        invalid_extension = ValidationResult.failure(
            [FileExtensionValidator.InvalidFileExtension(self, value)],
        )
        extension = Path(value).suffix.lstrip(".")
        if self.valid_extensions is not None and extension.upper() not in self.valid_extensions:
            return invalid_extension
        return self.success()


class AtomSelectionValidator(Validator):
    """Check that input is a valid selection string"""

    def __init__(
        self,
        failure_description: str | None = None,
    ) -> None:
        super().__init__(failure_description=failure_description)

    class InvalidAtomSelection(Failure):
        """Indicates that the selection string is invalid."""

    def validate(self, value: str) -> ValidationResult:
        """Validates that `value` is a valid selection strin (i.e. it returns an AtomGroup).

        Args:
            value: The value to validate.

        Returns:
            The result of the validation.
        """
        failure = AtomSelectionValidator.InvalidAtomSelection(self, value)
        invalid_selection = ValidationResult.failure([failure])
        value = value if value else "all"
        try:
            mda.core.selection.Parser.parse(selectstr=value, selgroups={})
        except mda.exceptions.SelectionError as e:
            failure.description = f"Invalid atom selection - {e}"
            return invalid_selection
        except mda.exceptions.NoDataError as e:
            failure.description = f"Invalid atom selection - {e}"
            return invalid_selection
        except AttributeError as e:
            failure.description = f"Invalid atom selection - {e}"
            return invalid_selection
        else:
            return self.success()



from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import (
    Button,
    Input,
)

from mda_tui.validators import IntegerOrNoneValidator


class MDARun(Horizontal):
    """Widget for selecting run parameters"""

    def compose(self) -> ComposeResult:
        """Create the layout for setting run parameters"""
        start = Input(
            placeholder="start",
            validators=IntegerOrNoneValidator(
                failure_description="'start' must be an integer or empty",
            ),
            id="start",
        )
        stop = Input(
            placeholder="stop",
            validators=IntegerOrNoneValidator(
                failure_description="'stop' must be an integer or empty",
            ),
            id="stop",
        )
        step = Input(
            placeholder="step",
            validators=IntegerOrNoneValidator(
                failure_description="'step' must be an integer or empty",
            ),
            id="step",
        )
        run = Button("run", id="run-button")

        # Define tooltips
        start.tooltip = "first frame to transform (0-based index). If empty, defaults to the first frame of the trajectory."
        stop.tooltip = "final frame to transform (non-inclusive). If empty, defaults to the final frame of the trajectory."
        step.tooltip = "number of frames to skip between each transformed frame. If empty, defaults to no frame skipped."
        run.tooltip = "run the transformation. This will read the input trajectory, apply the selected transformation to the selected frames and write the transformed trajecotry to the output file."

        yield start
        yield stop
        yield step
        yield run

    @on(Input.Changed)
    def show_invalid_reasons(self, event: Input.Changed) -> None:
        if not event.validation_result.is_valid:
            for failure in event.validation_result.failures:
                self.notify(failure.description, severity="error", timeout=3)
            return

    @property
    def slice(self) -> tuple[int | None, int | None, int | None]:  # noqa:  A003
        """Get the start, stop, and step values as integers or None"""

        start = self.query_one("#start", Input).value
        stop = self.query_one("#stop", Input).value
        step = self.query_one("#step", Input).value

        start = None if not start else int(start)
        stop = None if not stop else int(stop)
        step = None if not step else int(step)

        return start, stop, step



"""
Global pytest fixtures
"""

# Use this file if you need to share any fixtures
# across multiple modules
# More information at
# https://docs.pytest.org/en/stable/how-to/fixtures.html#scope-sharing-fixtures-across-classes-modules-packages-or-session

import MDAnalysis as mda
import numpy as np
import pytest

from mda_tui.__main__ import MDA


@pytest.fixture()
def app():
    return MDA()


@pytest.fixture(scope="session")
def universe_filenames(tmp_path_factory):
    """
    Write a dummy universe to file for testing later.

    Note, we use `tmp_path_factory` because this fixture requies `module` scope
    so we can read the file after the fixture has been created, and
    `tmp_path` has function-level scope.
    """
    n_atoms = 2
    n_frames = 2
    u = mda.Universe.empty(n_atoms, trajectory=True)
    coordinates = np.empty((n_frames, u.atoms.n_atoms, 3))
    coordinates[0] = np.asarray(
        [
            [2, 2, 2],
            [11, 11, 11],  # outside the unit cell
        ],
    )
    # both atoms move -3 in each dimension and cross periodic boundaries
    coordinates[1] = np.asarray(
        [
            [9, 9, 9],
            [-2, -2, -2],
        ],
    )
    u.load_new(coordinates, order="fac")
    u.add_bonds(values=[(0, 1)])
    dims = np.asarray(
        [
            [10, 10, 10, 90, 90, 90],
        ],
    )
    u.dimensions = dims

    tmp_pdb = tmp_path_factory.getbasetemp() / "dummy_universe.pdb"
    tmp_xtc = tmp_path_factory.getbasetemp() / "dummy_universe.xtc"
    u.atoms.write(tmp_pdb.as_posix())
    with mda.Writer(tmp_xtc.as_posix()) as f:
        for _ts in u.trajectory:
            u.dimensions = dims
            f.write(u.atoms)

    return tmp_pdb, tmp_xtc



import pathlib

import MDAnalysis as mda
from textual import on
from textual.app import App, ComposeResult
from textual.widgets import (
    Button,
    Header,
    Select,
    TabbedContent,
    TabPane,
)

from mda_tui.widgets import (
    MDARun,
    TopologyReaderSelector,
    TrajectoryReaderSelector,
    TrajectoryWriterSelector,
    TransformationSelector,
)


class MDA(App):
    """A TUI for MDAnalysis."""

    CSS_PATH = "mda-tui.tcss"

    def __init__(self):
        super().__init__()
        self.title = "MDA TUI"
        self.sub_title = "Trajectory transformations your terminal!"

    def compose(self) -> ComposeResult:
        """Set up the layout of the app"""
        yield Header()
        with TabbedContent():
            with TabPane("Universe"):
                yield TopologyReaderSelector()
                yield TrajectoryReaderSelector()
            with TabPane("Transformations"):
                yield TransformationSelector()
            with TabPane("Output"):
                yield TrajectoryWriterSelector()
        yield MDARun(id="run")

    @on(Button.Pressed, "TopologyReaderSelector #openFile")
    def open_topology_input(self) -> None:
        """Show the `FileOpen` dialog when the button is pushed."""
        topology_widget = self.query_one(TopologyReaderSelector)
        self.push_screen(
            topology_widget.launch_dialogue(),
            callback=topology_widget.show_selected_file,
        )

    @on(Button.Pressed, "TrajectoryReaderSelector #openFile")
    def open_trajectory_input(self) -> None:
        """Show the `FileOpen` dialog when the button is pushed."""
        trajectory_widget = self.query_one(TrajectoryReaderSelector)
        self.push_screen(
            trajectory_widget.launch_dialogue(),
            callback=trajectory_widget.show_selected_file,
        )

    @on(Button.Pressed, "TrajectoryWriterSelector #openFile")
    def open_trajectory_output(self) -> None:
        """Show the `FileOpen` dialog when the button is pushed."""
        trajectory_widget = self.query_one(TrajectoryWriterSelector)
        self.push_screen(
            trajectory_widget.launch_dialogue(),
            callback=trajectory_widget.show_selected_file,
        )

    @on(Button.Pressed, "MDARun #run-button")
    def run_transformation(self) -> None:
        """Perform the transformation"""

        # Validate input
        topology_selector = self.query_one(TopologyReaderSelector)
        valid_topology_result = topology_selector.validate()
        if not valid_topology_result.is_valid:
            for failure in valid_topology_result.failures:
                self.notify(failure.description, severity="error", timeout=10)
            return

        trajectory_selector = self.query_one(TrajectoryReaderSelector)
        valid_trajectory_result = trajectory_selector.validate()
        if not valid_trajectory_result.is_valid:
            for failure in valid_trajectory_result.failures:
                self.notify(failure.description, severity="error", timeout=10)
            return

        output_trajectory_selector = self.query_one(TrajectoryWriterSelector)
        valid_output_trajectory_result = output_trajectory_selector.validate()
        if not valid_output_trajectory_result.is_valid:
            for failure in valid_output_trajectory_result.failures:
                self.notify(failure.description, severity="error", timeout=10)
            return

        # Load universe
        topology = topology_selector.file
        trajectory = trajectory_selector.file
        u = mda.Universe(topology.as_posix(), trajectory.as_posix())

        # Apply transformation
        transformation_wrapper = (
            self.query_one(TransformationSelector).query_one("#transformation", Select).value
        )
        if transformation_wrapper is None:
            self.notify(
                "No transformation selected - please select a transformation",
                severity="error",
                timeout=20,
            )
            return
        valid_transformation_wrapper_results = transformation_wrapper.validate()
        all_valid = all(result.is_valid for result in valid_transformation_wrapper_results)
        if not all_valid:
            for result in valid_transformation_wrapper_results:
                for failure in result.failures:
                    self.notify(failure.description, severity="error", timeout=10)
            return
        transformation = transformation_wrapper.setup_transformation(universe=u)
        u.trajectory.add_transformations(transformation)

        # Write transformed trajectory
        (start, stop, step) = self.query_one(MDARun).slice
        output_trajectory: pathlib.Path = self.query_one(TrajectoryWriterSelector).file
        output_trajectory.parent.mkdir(parents=True, exist_ok=True)
        with mda.Writer(output_trajectory.as_posix()) as f:
            # TODO: show a progress bar
            for _ts in u.trajectory[start:stop:step]:
                f.write(u.atoms)

        self.notify(f"Finished writing transfomed trajectory to {output_trajectory}", timeout=20)


def main():
    app = MDA()
    app.run()


if __name__ == "__main__":
    main()



from MDAnalysis import Universe
from MDAnalysis.transformations import (
    NoJump as nojump,
    TransformationBase,
    center_in_box,
    translate,
    unwrap,
    wrap,
)
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import (
    ContentSwitcher,
    Input,
    Label,
    Select,
    Static,
    Switch,
)

from mda_tui.validators import AtomSelectionValidator, NumberOrNoneValidator


class WidgetWithLabel(Static):
    """Add a label to a widget"""

    def __init__(
        self,
        label: str,
        widget: Static,
        id: str | None = None,
        tooltip: str = "",
    ) -> None:
        super().__init__(id=id)
        self.label_text = label
        self.labelled_widget = widget
        self.tooltip = tooltip

    def compose(self) -> ComposeResult:
        yield Label(self.label_text)
        yield self.labelled_widget


class TransformationSelector(Vertical):
    """Widget for selecting a transformation to apply"""

    def compose(self) -> ComposeResult:
        """Create the layout for the transformation selector"""

        translate = Translate(id=Translate.id)
        center_in_box = CenterInBox(id=CenterInBox.id)
        wrap = Wrap(id=Wrap.id)
        unwrap = Unwrap(id=Unwrap.id)
        nojump = NoJump(id=NoJump.id)

        options = [
            (translate.description, translate),
            (center_in_box.description, center_in_box),
            (wrap.description, wrap),
            (unwrap.description, unwrap),
            (nojump.description, nojump),
        ]
        select: Select[TransformationBase] = Select(
            options=options,
            prompt="select transformation",
            id="transformation",
        )
        select.tooltip = "select a transformation to apply"
        yield select
        with ContentSwitcher(initial=None, id="selector"):
            yield translate
            yield center_in_box
            yield wrap
            yield unwrap
            yield nojump

    @on(Select.Changed, "#transformation")
    def select_transformation(self, event) -> None:
        selected_id = None if event.value is None else event.value.id
        self.query_one(ContentSwitcher).current = selected_id


class Translate(Vertical):
    """Widgets for setting parameters for translate transformation"""

    description = "Translate coordinates by a given vector"
    transformation = translate
    id = str(translate).removeprefix("<class '").removesuffix("'>")  # noqa: A003

    def compose(self) -> ComposeResult:
        """Create layout of parameter widgets"""

        vector = Horizontal(
            Input(
                placeholder="x",
                validators=NumberOrNoneValidator(
                    failure_description="'x' must be a number or empty",
                ),
                id="translate_x",
            ),
            Input(
                placeholder="y",
                validators=NumberOrNoneValidator(
                    failure_description="'y' must be a number or empty",
                ),
                id="translate_y",
            ),
            Input(
                placeholder="z",
                validators=NumberOrNoneValidator(
                    failure_description="'z' must be a number or empty",
                ),
                id="translate_z",
            ),
        )

        # Define tooltips
        vector_tooltip = "vector by which the all atoms will be translated"

        yield WidgetWithLabel(
            label="vector",
            widget=vector,
            id="translate_vector",
            tooltip=vector_tooltip,
        )

    @property
    def vector(self):
        x = self.query_one("#translate_x", Input).value
        y = self.query_one("#translate_y", Input).value
        z = self.query_one("#translate_z", Input).value
        x = float(x) if x else 0
        y = float(y) if y else 0
        z = float(z) if z else 0
        return [x, y, z]

    def setup_transformation(self, universe: Universe):  # noqa: ARG002
        """Initialise the transformation for a given universe"""
        return self.transformation(vector=self.vector)

    def validate(self):
        return [widget.validate(widget.value) for widget in self.query(Input)]


class CenterInBox(Vertical):
    """Widgets for setting parameters for center_in_box transformation"""

    description = "Center atoms / molecules"
    transformation = center_in_box
    id = str(center_in_box).removeprefix("<class '").removesuffix("'>")  # noqa: A003

    def compose(self) -> ComposeResult:
        """Create layout of parameter widgets"""

        ag = Input(
            placeholder="atom selection",
            validators=AtomSelectionValidator(
                failure_description="invalid atom selection",
            ),
            id="ag",
        )

        options = [
            ("centor of geometry", "geometry"),
            ("center of mass", "mass"),
        ]
        method: Select[str] = Select(
            options=options,
            prompt="select centering method",
            value="geometry",
            id="centering_method",
        )

        point = Horizontal(
            Input(
                placeholder="x",
                validators=NumberOrNoneValidator(
                    failure_description="'x' must be a number or empty",
                ),
                id="center_x",
            ),
            Input(
                placeholder="y",
                validators=NumberOrNoneValidator(
                    failure_description="'y' must be a number or empty",
                ),
                id="center_y",
            ),
            Input(
                placeholder="z",
                validators=NumberOrNoneValidator(
                    failure_description="'z' must be a number or empty",
                ),
                id="center_z",
            ),
        )
        wrap = Switch()

        # Define tooltips
        ag_tooltip = "selection string for atom group to be centered on the unit cell. If empty, will center all atoms."
        method_tooltip = "choose the method of centering on the selected AtomGroup."
        point_tooltip = "if provided, overrides the unit cell center - the coordinates of each timestep are translated so that the center of mass/geometry of the selected AtomGroup is aligned to this position instead. Leave empty to translate to the center of the unit cell."
        wrap_tooltip = "if enabled, the atoms from the selected AtomGroup will be moved to the primary unit cell before calculating the center of mass or geometry."

        yield WidgetWithLabel(label="ag", widget=ag, id="center_ag", tooltip=ag_tooltip)
        yield WidgetWithLabel(
            label="center",
            widget=method,
            id="center_method",
            tooltip=method_tooltip,
        )
        yield WidgetWithLabel(label="point", widget=point, id="center_on", tooltip=point_tooltip)
        yield WidgetWithLabel(label="wrap", widget=wrap, id="center_wrap", tooltip=wrap_tooltip)

    @property
    def selection(self):
        sel = self.query_one("#ag", Input).value
        sel = sel if sel else "all"
        return sel

    @property
    def method(self):
        return self.query_one(Select).value

    @property
    def point(self):
        x = self.query_one("#center_x", Input).value
        y = self.query_one("#center_y", Input).value
        z = self.query_one("#center_z", Input).value
        x = float(x) if x else None
        y = float(y) if y else None
        z = float(z) if z else None
        return None if None in [x, y, z] else [x, y, z]

    @property
    def wrap(self):
        return self.query_one(Switch).value

    def setup_transformation(self, universe: Universe):
        """Initialise the transformation for a given universe"""
        ag = universe.select_atoms(self.selection)
        return self.transformation(ag=ag, center=self.method, point=self.point, wrap=self.wrap)

    def validate(self):
        return [widget.validate(widget.value) for widget in self.query(Input)]


class Wrap(Vertical):
    """Widgets for setting parameters for the wrap transformation"""

    description = "Wrap atoms into the unit cell"
    transformation = wrap
    id = str(wrap).removeprefix("<class '").removesuffix("'>")  # noqa: A003

    """
    ag: Atomgroup
        Atomgroup to be wrapped in the unit cell
    compound : {'atoms', 'group', 'residues', 'segments', 'fragments'}, optional
        The group which will be kept together through the shifting process.
    """

    def compose(self) -> ComposeResult:
        """Create layout of parameter widgets"""

        ag = Input(
            placeholder="atom selection",
            validators=AtomSelectionValidator(
                failure_description="invalid atom selection",
            ),
            id="ag",
        )

        options = [
            ("individual atoms", "atoms"),
            ("selected atom group", "group"),
            ("residues", "residues"),
            ("segments", "segments"),
            ("fragments", "fragments"),
        ]
        compound: Select[str] = Select(
            options=options,
            prompt="select compound to wrap",
            value="atoms",
            id="wrap_compound",
        )

        # define tooltips
        ag_tooltip = "selection string for atom group to be wrapped into the unit cell. If empty, all atoms will be wrapped."
        compound_tooltip = (
            "select the group that will be kept together through the wrapping transformation"
        )

        yield WidgetWithLabel(label="ag", widget=ag, id="wrap_ag", tooltip=ag_tooltip)
        yield WidgetWithLabel(
            label="compound",
            widget=compound,
            id="wrap_compound",
            tooltip=compound_tooltip,
        )

    @property
    def selection(self):
        sel = self.query_one("#ag", Input).value
        sel = sel if sel else "all"
        return sel

    @property
    def compound(self):
        return self.query_one(Select).value

    def setup_transformation(self, universe: Universe):
        """Initialise the transformation for a given universe"""
        ag = universe.select_atoms(self.selection)
        return self.transformation(ag=ag, compound=self.compound)

    def validate(self):
        return [widget.validate(widget.value) for widget in self.query(Input)]


class Unwrap(Vertical):
    """Widgets for setting parameters for the unwrap transformation"""

    description = "Unwrap an atom group"
    transformation = unwrap
    id = str(unwrap).removeprefix("<class '").removesuffix("'>")  # noqa: A003

    def compose(self) -> ComposeResult:
        """Create layout of parameter widgets"""

        ag = Input(
            placeholder="atom selection",
            validators=AtomSelectionValidator(
                failure_description="invalid atom selection",
            ),
            id="ag",
        )

        # define tooltips
        ag_tooltip = "selection string for atom group to be unwrapped into the unit cell. If empty, all fragments will be unwrapped."

        yield WidgetWithLabel(label="ag", widget=ag, id="unwrap_ag", tooltip=ag_tooltip)

    @property
    def selection(self):
        sel = self.query_one("#ag", Input).value
        sel = sel if sel else "all"
        return sel

    def setup_transformation(self, universe: Universe):
        """Initialise the transformation for a given universe"""
        ag = universe.select_atoms(self.selection)
        return self.transformation(ag=ag)

    def validate(self):
        return [widget.validate(widget.value) for widget in self.query(Input)]


class NoJump(Vertical):
    """Widgets for setting parameters for nojump transformation"""

    description = "Prevent jumps across periodic boundaries"
    transformation = nojump
    id = str(nojump).removeprefix("<class '").removesuffix("'>")  # noqa: A003

    def compose(self) -> ComposeResult:
        """Create layout of parameter widgets"""

        check_continuity = Switch(value=True)

        # Define tooltips
        check_continuity_tooltip = "check the trajectory step size is not more than 1"

        yield WidgetWithLabel(
            label="check continuity",
            widget=check_continuity,
            id="check_continuity",
            tooltip=check_continuity_tooltip,
        )

    @property
    def check_continuity(self):
        return self.query_one(Switch).value

    def setup_transformation(self, universe: Universe):  # noqa: ARG002
        """Initialise the transformation for a given universe"""
        return self.transformation(check_continuity=self.check_continuity)

    def validate(self):
        return [widget.validate(widget.value) for widget in self.query(Input)]



from pathlib import Path
from typing import Callable, ClassVar

import MDAnalysis as mda
from textual.app import ComposeResult
from textual.containers import (
    Horizontal,
)
from textual.validation import Validator
from textual.widgets import (
    Button,
    Input,
)
import textual_fspicker
from typing_extensions import TypeAlias

from mda_tui.validators import (
    FileExtensionValidator,
    FileValidator,
)
from mda_tui.widgets.file_dialogues import FileOpen, FileSave

FilterFunction: TypeAlias = Callable[[Path], bool]
FileFilters: TypeAlias = list[tuple[str, FilterFunction]]


class FileSelector(Horizontal):
    """Widget for selecting a file using a dialogue box or Input widget

    Attributes:
        filters: List of filters to use when selecting files
        input_placeholder: String to show in empty Input widget
        input_id: Textual ID to use for the Input widget
        button_id: Textual ID to use for the Button widget
    """

    filters: FileFilters | None = None
    placeholder: str = "select file"
    input_id: str | None = None
    button_id: str | None = "openFile"
    validators: Validator | list[Validator] | None = None

    def compose(self) -> ComposeResult:
        """Create the layout for the file selector"""
        yield Button("browse", id=self.button_id)
        yield Input(placeholder=self.placeholder, validators=self.validators, id=self.input_id)

    @classmethod
    def create_file_filters(cls, extensions: list[str]) -> FileFilters:
        """Create a list of file filters from on a list of file extensions.

        This is a naive approach to creating filters - each filter will show a single file
        type (with the same extension as the name of the filter)

        e.g. the 'PSF' filter will show all files with a '.psf' extension.

        The exception to this is the 'ANY' filter, which shows all files.
        """
        filters = [
            (extension, lambda file, e=extension: file.suffix.lower() == f".{e.lower()}")
            for extension in extensions
        ]
        return [("Any", lambda _: True), *filters]

    def launch_dialogue(self):
        """Open a file dialogue box"""
        return FileOpen(
            ".",
            filters=textual_fspicker.Filters(*self.filters),
        )

    def show_selected_file(self, path):
        """Display the selected file in the Input widget"""
        self.query_one(Input).value = "" if path is None else str(path)

    @property
    def file(self) -> Path | None:
        return Path(self.query_one(Input).value)

    def validate(self):
        input = self.query_one(Input)  # noqa: A001
        return input.validate(input.value)


class TopologyReaderSelector(FileSelector):
    """Widget for selecting a topology to load"""

    filters: FileFilters = FileSelector.create_file_filters(
        extensions=sorted(mda._PARSERS.keys()),
    )
    placeholder: str = "select topology file"
    validators: ClassVar = [
        FileValidator(failure_description="Input topology file does not exist"),
        FileExtensionValidator(
            valid_extensions=sorted(mda._PARSERS.keys()),
            failure_description="Unknown input topology format",
        ),
    ]
    tooltip = "select a valid topology file to load."


class TrajectoryReaderSelector(FileSelector):
    """Widget for selecting a trajectory to load"""

    filters: FileFilters = FileSelector.create_file_filters(
        extensions=sorted(mda._READERS.keys()),
    )
    placeholder: str = "select trajectory file"
    validators: ClassVar = [
        FileValidator(failure_description="Input trajectory file does not exist"),
        FileExtensionValidator(
            valid_extensions=sorted(mda._READERS.keys()),
            failure_description="Unknown input trajectory format",
        ),
    ]
    tooltip = "select a valid trajectory file to load."


class TrajectoryWriterSelector(FileSelector):
    """Widget for selecting a trajectory to write"""

    filters: FileFilters = FileSelector.create_file_filters(
        extensions=sorted(mda._MULTIFRAME_WRITERS.keys()),
    )
    placeholder: str = "select transformed trajectory file"
    validators: ClassVar = [
        FileExtensionValidator(
            valid_extensions=sorted(mda._MULTIFRAME_WRITERS.keys()),
            failure_description="Unknown output trajectory format",
        ),
    ]
    tooltip = "select file to write the transformed trajetory to. Any existing file of the same name will be overwritten."

    def launch_dialogue(self):
        """Open a file dialogue box"""
        return FileSave(
            location=".",
            filters=textual_fspicker.Filters(*self.filters),
        )



import pathlib

import MDAnalysis as mda
import numpy as np
from numpy.testing import assert_allclose
from packaging.specifiers import SpecifierSet
import pytest
from textual.widgets import (
    Input,
    Select,
    Switch,
)

from mda_tui.widgets import (
    TopologyReaderSelector,
    TrajectoryReaderSelector,
    TrajectoryWriterSelector,
    TransformationSelector,
    transformations,
)

mda_supported_versions = SpecifierSet(">2.6.1", prereleases=True)


@pytest.mark.asyncio()
async def test_translate(app, universe_filenames: tuple[pathlib.Path, pathlib.Path]):
    pdb, xtc = universe_filenames
    xtc_output = xtc.parent / "translated_trajectory.xtc"
    translate_by = 10

    async with app.run_test() as pilot:
        topology_reader_input = pilot.app.query_one(TopologyReaderSelector).query_one(Input)
        trajectory_reader_input = pilot.app.query_one(TrajectoryReaderSelector).query_one(Input)
        trajectory_writer_output = pilot.app.query_one(TrajectoryWriterSelector).query_one(Input)
        transformation_selector_input = pilot.app.query_one(TransformationSelector).query_one(
            "#transformation",
            Select,
        )
        translate_transformation = pilot.app.query_one(transformations.Translate)

        topology_reader_input.value = pdb.as_posix()
        trajectory_reader_input.value = xtc.as_posix()
        trajectory_writer_output.value = xtc_output.as_posix()
        transformation_selector_input.value = translate_transformation
        translate_transformation.query_one("#translate_x", Input).value = str(translate_by)
        translate_transformation.query_one("#translate_y", Input).value = str(translate_by)
        translate_transformation.query_one("#translate_z", Input).value = str(translate_by)

        pilot.app.run_transformation()

    u = mda.Universe(pdb.as_posix(), xtc.as_posix())
    u_translated = mda.Universe(pdb.as_posix(), xtc_output.as_posix())
    diff = (u.trajectory.timeseries() - u_translated.trajectory.timeseries()).flatten()
    assert_allclose(np.abs(diff), translate_by, atol=1e-5)


@pytest.mark.asyncio()
async def test_center_in_box(app, universe_filenames: tuple[pathlib.Path, pathlib.Path]):
    pdb, xtc = universe_filenames
    xtc_output = xtc.parent / "centered_trajectory.xtc"
    ag = "all"

    async with app.run_test() as pilot:
        topology_reader_input = pilot.app.query_one(TopologyReaderSelector).query_one(Input)
        trajectory_reader_input = pilot.app.query_one(TrajectoryReaderSelector).query_one(Input)
        trajectory_writer_output = pilot.app.query_one(TrajectoryWriterSelector).query_one(Input)
        transformation_selector_input = pilot.app.query_one(TransformationSelector).query_one(
            "#transformation",
            Select,
        )
        center_transformation = pilot.app.query_one(transformations.CenterInBox)

        topology_reader_input.value = pdb.as_posix()
        trajectory_reader_input.value = xtc.as_posix()
        trajectory_writer_output.value = xtc_output.as_posix()
        transformation_selector_input.value = center_transformation
        center_transformation.query_one("#ag", Input).value = ag

        pilot.app.run_transformation()

    u = mda.Universe(pdb.as_posix(), xtc.as_posix())
    u_centered = mda.Universe(pdb.as_posix(), xtc_output.as_posix())
    box = u.dimensions[:3]
    with pytest.raises(AssertionError):
        assert_allclose(box / 2, np.mean(u.atoms.positions, axis=0))
    assert_allclose(box / 2, np.mean(u_centered.atoms.positions, axis=0))


@pytest.mark.asyncio()
async def test_wrap(app, universe_filenames: tuple[pathlib.Path, pathlib.Path]):
    pdb, xtc = universe_filenames
    xtc_output = xtc.parent / "wrapped_trajectory.xtc"
    ag = "all"

    async with app.run_test() as pilot:
        topology_reader_input = pilot.app.query_one(TopologyReaderSelector).query_one(Input)
        trajectory_reader_input = pilot.app.query_one(TrajectoryReaderSelector).query_one(Input)
        trajectory_writer_output = pilot.app.query_one(TrajectoryWriterSelector).query_one(Input)
        transformation_selector_input = pilot.app.query_one(TransformationSelector).query_one(
            "#transformation",
            Select,
        )
        wrap_transformation = pilot.app.query_one(transformations.Wrap)

        topology_reader_input.value = pdb.as_posix()
        trajectory_reader_input.value = xtc.as_posix()
        trajectory_writer_output.value = xtc_output.as_posix()
        transformation_selector_input.value = wrap_transformation
        wrap_transformation.query_one("#ag", Input).value = ag

        pilot.app.run_transformation()

    u = mda.Universe(pdb.as_posix(), xtc.as_posix())
    u_wrapped = mda.Universe(pdb.as_posix(), xtc_output.as_posix())
    with pytest.raises(AssertionError):
        assert_allclose(u.atoms.wrap(inplace=False), u.atoms.positions)
    assert_allclose(u.atoms.wrap(inplace=False), u_wrapped.atoms.positions)


@pytest.mark.asyncio()
async def test_unwrap(app, universe_filenames: tuple[pathlib.Path, pathlib.Path]):
    pdb, xtc = universe_filenames
    xtc_output = xtc.parent / "unwrapped_trajectory.xtc"
    ag = "all"

    async with app.run_test() as pilot:
        topology_reader_input = pilot.app.query_one(TopologyReaderSelector).query_one(Input)
        trajectory_reader_input = pilot.app.query_one(TrajectoryReaderSelector).query_one(Input)
        trajectory_writer_output = pilot.app.query_one(TrajectoryWriterSelector).query_one(Input)
        transformation_selector_input = pilot.app.query_one(TransformationSelector).query_one(
            "#transformation",
            Select,
        )
        unwrap_transformation = pilot.app.query_one(transformations.Unwrap)

        topology_reader_input.value = pdb.as_posix()
        trajectory_reader_input.value = xtc.as_posix()
        trajectory_writer_output.value = xtc_output.as_posix()
        transformation_selector_input.value = unwrap_transformation
        unwrap_transformation.query_one("#ag", Input).value = ag

        pilot.app.run_transformation()

    u = mda.Universe(pdb.as_posix(), xtc.as_posix())
    atom_1_position, atom_2_position = u.bonds[0].atoms.positions
    with pytest.raises(AssertionError):
        assert all(np.abs(atom_1_position - atom_2_position) < (u.dimensions[:3] / 2))

    u_unwrapped = mda.Universe(pdb.as_posix(), xtc_output.as_posix())
    atom_1_position, atom_2_position = u_unwrapped.bonds[0].atoms.positions
    assert all(np.abs(atom_1_position - atom_2_position) < (u_unwrapped.dimensions[:3] / 2))


@pytest.mark.xfail(
    mda.__version__ not in mda_supported_versions,
    reason="bug in NoJump transformation is fixed in MDAnalysis >2.6.1",
    strict=True,
)
@pytest.mark.asyncio()
async def test_nojump(app, universe_filenames: tuple[pathlib.Path, pathlib.Path]):
    pdb, xtc = universe_filenames
    xtc_output = xtc.parent / "nojump_trajectory.xtc"

    async with app.run_test() as pilot:
        topology_reader_input = pilot.app.query_one(TopologyReaderSelector).query_one(Input)
        trajectory_reader_input = pilot.app.query_one(TrajectoryReaderSelector).query_one(Input)
        trajectory_writer_output = pilot.app.query_one(TrajectoryWriterSelector).query_one(Input)
        transformation_selector_input = pilot.app.query_one(TransformationSelector).query_one(
            "#transformation",
            Select,
        )
        nojump_transformation = pilot.app.query_one(transformations.NoJump)
        nojump_transformation.query_one(Switch).value = True

        topology_reader_input.value = pdb.as_posix()
        trajectory_reader_input.value = xtc.as_posix()
        trajectory_writer_output.value = xtc_output.as_posix()
        transformation_selector_input.value = nojump_transformation

        pilot.app.run_transformation()

    u = mda.Universe(pdb.as_posix(), xtc.as_posix())
    distances_moved = np.diff(u.trajectory.timeseries(), axis=0)
    with pytest.raises(AssertionError):
        assert np.all(np.abs(distances_moved) < (u.dimensions[:3] / 2))

    u_nojump = mda.Universe(pdb.as_posix(), xtc_output.as_posix())
    distances_moved = np.diff(u_nojump.trajectory.timeseries(), axis=0)
    assert np.all(np.abs(distances_moved) < (u_nojump.dimensions[:3] / 2))



import pathlib

from textual_fspicker.file_dialog import BaseFileDialog
from textual_fspicker.path_filters import Filters


class FileOpen(BaseFileDialog):
    """A file opening dialog.

    This is the same as textual_fspicker.FileOpen except the 'select_button' text
    is changed from 'Open' to 'Select'.
    """

    def __init__(
        self,
        location: str | pathlib.Path = ".",
        title: str = "Open",
        *,
        select_button="Select",
        filters: Filters | None = None,
        must_exist: bool = True,
    ) -> None:
        """Initialise the `FileOpen` dialog.

        Args:
            location: Optional starting location.
            title: Optional title.
            filters: Optional filters to show in the dialog.
            must_exist: Flag to say if the file must exist.
        """
        super().__init__(location, title, select_button=select_button, filters=filters)
        self._must_exist = must_exist
        """Must the file exist?"""

    def _should_return(self, candidate: pathlib.Path) -> bool:
        """Perform the final checks on the chosen file.

        Args:
            candidate: The file to check.
        """
        if self._must_exist and not candidate.exists():
            self._set_error("The file must exist")
            return False
        return True


class FileSave(BaseFileDialog):
    """A file save dialog.

    This is the same as textual_fspicker.FileSave except the 'select_button' text
    is changed from 'Save' to 'Select'.
    """

    def __init__(
        self,
        location: str | pathlib.Path = ".",
        title: str = "Save as",
        *,
        filters: Filters | None = None,
        can_overwrite: bool = True,
        select_button="Select",
    ) -> None:
        """Initialise the `FileOpen` dialog.

        Args:
            location: Optional starting location.
            title: Optional title.
            filters: Optional filters to show in the dialog.
            can_overwrite: Flag to say if an existing file can be overwritten.
        """
        super().__init__(location, title, select_button=select_button, filters=filters)
        self._can_overwrite = can_overwrite
        """Can an existing file be overwritten?"""

    def _should_return(self, candidate: pathlib.Path) -> bool:
        """Perform the final checks on the chosen file.

        Args:
            candidate: The file to check.
        """
        if candidate.exists() and not self._can_overwrite:
            self._set_error("Overwrite is not allowed")
            return False
        return True
